#!/usr/bin/env python3

# avg_2hu.py
# version 6
#
# Author: Intel A80486DX2-66
# License: Creative Commons Zero 1.0 Universal or Unlicense

from decimal import Context, Decimal, getcontext as decimal_getcontext
from math import trunc
from statistics import stdev
from sys import argv
from typing import List, Optional, Union
import re

MAX_TABLE_CELL_NUMBER_WIDTH = 11

decimal_getcontext().prec = MAX_TABLE_CELL_NUMBER_WIDTH - len(".")

None_filter = lambda l: [x for x in l if x is not None]
sum_or_skip = lambda l: sum(None_filter(l))
len_or_skip = lambda l: len(None_filter(l))
arithmetic_mean = lambda l: sum_or_skip(l) / len_or_skip(l)
standard_deviation = lambda l: stdev(None_filter(l))
strip_every = lambda l: [s.strip() for s in l]
is_float_regex = re.compile(r"^\d*\.?\d*?$")
dec_or_None = lambda s: Decimal(s) if len(s) > 0 and is_float_regex.match(s) \
                        else None
remove_notes = lambda l: [s.split('[')[0] for s in l]
select_column = lambda l, i: [row[i] for row in l]
remove_newlines = lambda s: s.replace('\r', '').replace('\n', '')
cr_lf = lambda lines: [remove_newlines(line) + '\r\n' for line in lines]

def dec_int_part_len(x: Decimal) -> str:
    return len(str(trunc(x)))


def dec_str_round_exact(x: Decimal, n_digits: int,
                        context: Context = decimal_getcontext()) -> str:
    round_to = context.prec if n_digits > context.prec else n_digits
    total_chars = dec_int_part_len(x) + 1 + n_digits

    result_string = str(round(x, round_to))

    return result_string[:total_chars].ljust(total_chars, '0')


def table_cell_fmt(x: Union[int, float, Decimal],
                   max_width: int = MAX_TABLE_CELL_NUMBER_WIDTH) -> str:
    string_float = str(x)
    n_digits = max(len(string_float) - len('.') - dec_int_part_len(x), 0)
    result_string = dec_str_round_exact(x, n_digits)

    if len(result_string) > max_width:
        raise OverflowError(f"The value x={str(x)} doesn't fit in {max_width=} "
                            "characters")

    return result_string[:max_width].ljust(max_width)


def parse_line(line: str) -> List[Optional[Decimal]]:
    height, weight = remove_notes(strip_every(line.split('|')[3:5]))

    height = dec_or_None(height)
    weight = dec_or_None(weight)

    return height, weight


BAR_SEPARATOR = "+------+------------------------------+-------------" \
                "+-------------+"

global line_to_insert_at

line_to_insert_at = {
    "avg": None,
    "stdev": None
}


def gather_info_from_file(file_path: str) -> List[Optional[Decimal]]:
    global line_to_insert_at

    line_to_insert_at["avg"] = 1  # separator offset

    reading_step = 0

    with open(file_path) as file:
        for raw_line in file:
            line = raw_line.strip()
            if line == BAR_SEPARATOR:
                reading_step += 1
            else:
                if reading_step == 2:
                    yield parse_line(line)

            if reading_step <= 2:
                line_to_insert_at["avg"] += 1  # counter

    line_to_insert_at["stdev"] = line_to_insert_at["avg"] + 1


ARGS = {
    "memory_optimization": ["-Om"],
    "speed_optimization": ["-Os"]
}

if __name__ == "__main__":
    assert len(argv) > 1, "Please specify the chart file as the first argument"

    avg_height = None
    avg_weight = None
    stdev_height = None
    stdev_weight = None

    print("Preparing data")

    input_data = gather_info_from_file(argv[1])

    print("Processing")
    if len(argv) <= 2 or argv[2] in ARGS["memory_optimization"]:
        sum_height = Decimal('0.0')
        sum_height_squared = Decimal('0.0')
        sum_weight = Decimal('0.0')
        sum_weight_squared = Decimal('0.0')
        length_height = Decimal(0)
        length_weight = Decimal(0)

        for height, weight in input_data:
            if height is not None:
                sum_height += height
                sum_height_squared += height**2
                length_height += 1
            if weight is not None:
                sum_weight += weight
                sum_weight_squared += weight**2
                length_weight += 1

        if length_height == 0 or length_weight == 0:
            raise SystemExit("No data to process")

        avg_height = sum_height / length_height
        avg_weight = sum_weight / length_weight

        # Compute sample standard deviation
        mean_squared_height = sum_height_squared / length_height
        mean_squared_weight = sum_weight_squared / length_weight

        calculate_variance = \
            lambda mean_squared, avg, n: (mean_squared - avg**2) * (n / (n - 1))
        variance_height = \
            calculate_variance(mean_squared_height, avg_height, length_height)
        variance_weight = \
            calculate_variance(mean_squared_weight, avg_weight, length_weight)

        assert variance_height >= Decimal(0), \
            "Variance for height is negative, cannot compute standard " \
            "deviation."
        assert variance_weight >= Decimal(0), \
            "Variance for weight is negative, cannot compute standard " \
            "deviation."

        stdev_height = variance_height.sqrt()
        stdev_weight = variance_weight.sqrt()
    elif len(argv) > 2 and argv[2] in ARGS["speed_optimization"]:
        gathered = list(input_data)

        heights = select_column(gathered, 0)
        weights = select_column(gathered, 1)

        avg_height = arithmetic_mean(heights)
        avg_weight = arithmetic_mean(weights)

        # Compute sample standard deviation
        stdev_height = standard_deviation(heights)
        stdev_weight = standard_deviation(weights)

    avg_height_str = table_cell_fmt(avg_height)
    avg_weight_str = table_cell_fmt(avg_weight)
    stdev_height_str = table_cell_fmt(stdev_height)
    stdev_weight_str = table_cell_fmt(stdev_weight)

    print(f"Arithmetic mean: height = {avg_height}, weight = {avg_weight}")
    print(f"Standard deviation: height = {stdev_height}, weight = "
          f"weight = {stdev_weight}")

    print("Updating the file")

    with open(argv[1]) as file:
        lines = file.readlines()

    lines[line_to_insert_at["avg"]] = \
        f"| avg. | *** Arithmetic means[-1]     | {avg_height_str} | " \
        f"{avg_weight_str} |\n"
    lines[line_to_insert_at["stdev"]] = \
        f"| sdv. | *** Standard deviations[-1]  | {stdev_height_str} | " \
        f"{stdev_weight_str} |\n"

    with open(argv[1], "w") as file:
        file.writelines(cr_lf(lines))

    print("All done")
